@@ -23,29 +23,31 @@
 #include <libgnome/gnome-url.h>
 #include <libgnomevfs/gnome-vfs.h>
 #include <libxml/xmlIO.h>
-
+#include <libxslt/xslt.h>
+#include <libxslt/xsltInternals.h>
+#include <libxslt/transform.h>
+#include <libxslt/xsltutils.h>
 #include <string.h>
 
-#include "yelp-db2html.h"
+#include "yelp-cache.h"
 #include "yelp-marshal.h"
 #include "yelp-reader.h"
 
-#define d(x) 
+#define d(x)
+
 #define BUFFER_SIZE 16384
 
-#define STAMP_MUTEX_LOCK    if(priv->async)g_mutex_lock(priv->stamp_mutex);
-#define STAMP_MUTEX_UNLOCK  if(priv->async)g_mutex_unlock(priv->stamp_mutex);
+#define STAMP_MUTEX_LOCK    g_mutex_lock(priv->stamp_mutex);
+#define STAMP_MUTEX_UNLOCK  g_mutex_unlock(priv->stamp_mutex);
 
 struct _YelpReaderPriv {
-	gboolean     async;
+	gint            stamp;
 
-	gint         stamp;
-
-        gboolean     active;
+        gboolean        active;
 
 	/* Locks */
-	GMutex      *stamp_mutex;
-	GAsyncQueue *thread_queue;
+	GMutex         *stamp_mutex;
+	GAsyncQueue    *thread_queue;
 };
 
 typedef struct {
@@ -68,16 +70,10 @@ typedef struct {
 	ReaderQueueType  type;
 } ReaderQueueData;
 
-
 static void      reader_class_init        (YelpReaderClass     *klass);
 static void      reader_init              (YelpReader          *reader);
 
-static void      reader_db_start          (ReaderThreadData    *th_data);
-static gint      reader_db_write          (ReaderThreadData    *th_data,
-					   const gchar         *buffer,
-					   gint                 len);
-static gint      reader_db_close          (ReaderThreadData    *th_data);
-static void      reader_man_info_start    (ReaderThreadData    *th_data);
+static void      reader_convert_start     (ReaderThreadData    *th_data);
 
 static void      reader_file_start        (ReaderThreadData    *th_data);
 
@@ -97,7 +93,8 @@ static void      reader_q_data_free       (ReaderQueueData     *q_data);
 /* FIXME: Solve this so we don't leak */
 static void      reader_th_data_free      (ReaderThreadData    *th_data);
 #endif
-
+static gchar *   reader_get_chunk         (const gchar         *document,
+					   const gchar         *section);
 
 enum {
 	START,
@@ -169,9 +166,9 @@ reader_class_init (YelpReaderClass *klass)
 			      G_STRUCT_OFFSET (YelpReaderClass,
 					       finished),
 			      NULL, NULL,
-			      yelp_marshal_VOID__VOID,
+			      yelp_marshal_VOID__POINTER,
 			      G_TYPE_NONE,
-			      0);
+			      1, G_TYPE_POINTER);
 	signals[CANCELLED] =
 		g_signal_new ("cancelled",
 			      G_TYPE_FROM_CLASS (klass),
@@ -209,125 +206,7 @@ reader_init (YelpReader *reader)
 }
 
 static void
-reader_db_start (ReaderThreadData *th_data)
-{
-	YelpReader         *reader;
-	YelpReaderPriv     *priv;
-	xmlOutputBufferPtr  buf;
-	GError             *error = NULL;
-	ReaderQueueData    *q_data;
-
-	g_return_if_fail (th_data != NULL);
-
-	reader = th_data->reader;
-	priv   = reader->priv;
-	
-	STAMP_MUTEX_LOCK;
-	
-	if (reader_check_cancelled (reader, th_data->stamp)) {
-		
-		STAMP_MUTEX_UNLOCK;
-		
-		return;
-	}
-
-	STAMP_MUTEX_UNLOCK;
-
-	buf = xmlAllocOutputBuffer (NULL);
-	
-	buf->writecallback = (xmlOutputWriteCallback) reader_db_write;
-	buf->closecallback = (xmlOutputCloseCallback) reader_db_close;
-	buf->context       = th_data;
-
-	q_data = reader_q_data_new (reader, th_data->stamp, 
-				    READER_QUEUE_TYPE_START);
-
-	g_async_queue_push (priv->thread_queue, q_data);
-	
-	yelp_db2html_convert (th_data->uri, buf, &error);
-	
-	if (error) {
-		g_warning ("Have an error here: %s\n", error->message);
-		/* FIXME: Fix error */
-/* 		g_signal_emit (reader, signals[ERROR], 0, error); */
-		g_error_free (error);
-	}
-
-        xmlOutputBufferClose (buf);
-}
-
-static int
-reader_db_write (ReaderThreadData *th_data, const gchar *buffer, gint len)
-{
-	YelpReader         *reader;
-	YelpReaderPriv     *priv;
-	ReaderQueueData    *q_data;
-
-	g_return_val_if_fail (th_data != NULL, -1);
-	
-	reader = th_data->reader;
-	priv   = reader->priv;
-
-	d(g_print ("reader_db_write: %d\n", len));
-	
-	STAMP_MUTEX_LOCK;
-	
-	if (reader_check_cancelled (reader, th_data->stamp)) {
-		
-		STAMP_MUTEX_UNLOCK;
-		
-		return -1;
-	}
-
-	STAMP_MUTEX_UNLOCK;
-
-	if (len <= 0) {
-		return 0;
-	}
-
-	q_data = reader_q_data_new (reader, th_data->stamp,
-				    READER_QUEUE_TYPE_DATA);
-
-	q_data->data = g_strndup (buffer, len);
-	
-	g_async_queue_push (priv->thread_queue, q_data);
-
-	return len;
-}
-
-static int
-reader_db_close (ReaderThreadData *th_data)
-{
-	YelpReader         *reader;
-	YelpReaderPriv     *priv;
-	ReaderQueueData    *q_data;
-
-	g_return_val_if_fail (th_data != NULL, -1);
-	
-	reader = th_data->reader;
-	priv   = reader->priv;
-
-	STAMP_MUTEX_LOCK;
-	
-	if (reader_check_cancelled (reader, th_data->stamp)) {
-		
-		STAMP_MUTEX_UNLOCK;
-		
-		return -1;
-	}
-
-	STAMP_MUTEX_UNLOCK;
-
-	q_data = reader_q_data_new (reader, th_data->stamp, 
-				    READER_QUEUE_TYPE_FINISHED);
-
-	g_async_queue_push (priv->thread_queue, q_data);
-
-	return 0;
-}
-
-static void
-reader_man_info_start (ReaderThreadData *th_data)
+reader_convert_start (ReaderThreadData *th_data)
 {
 	YelpReader      *reader;
 	YelpReaderPriv  *priv;
@@ -336,7 +215,7 @@ reader_man_info_start (ReaderThreadData *th_data)
 	GError          *error = NULL;
 	gint             exit_status;
 	ReaderQueueData *q_data;
-	gint             stamp;
+	gint             stamp; 
 	
 	g_return_if_fail (th_data != NULL);
 
@@ -345,7 +224,7 @@ reader_man_info_start (ReaderThreadData *th_data)
 	uri    = th_data->uri;
 	stamp  = th_data->stamp;
 	
-	d(g_print ("man_info_start\n"));
+	d(g_print ("convert_start\n"));
 
 	STAMP_MUTEX_LOCK;
 	
@@ -364,7 +243,21 @@ reader_man_info_start (ReaderThreadData *th_data)
 						yelp_uri_get_path (uri));
 		break;
 	case YELP_URI_TYPE_INFO:
-		command_line = g_strdup_printf ("gnome2-info2html %s",
+		if (yelp_uri_get_section (uri)) {
+			command_line = 
+				g_strdup_printf ("gnome2-info2html %s?%s",
+						 yelp_uri_get_path (uri),
+						 yelp_uri_get_section (uri));
+		} else { 
+			command_line = 
+				g_strdup_printf ("gnome2-info2html %s",
+						 yelp_uri_get_path (uri));
+		}
+		
+		break;
+	case YELP_URI_TYPE_DOCBOOK_XML:
+	case YELP_URI_TYPE_DOCBOOK_SGML:
+		command_line = g_strdup_printf ("yelp-db2html %s",
 						yelp_uri_get_path (uri));
 		break;
 	default:
@@ -385,6 +278,10 @@ reader_man_info_start (ReaderThreadData *th_data)
 				   NULL,
 				   &exit_status,
 				   &error /* FIXME */);
+	if (yelp_uri_get_type (uri) == YELP_URI_TYPE_DOCBOOK_XML ||
+	    yelp_uri_get_type (uri) == YELP_URI_TYPE_DOCBOOK_SGML) {
+		 yelp_cache_add (yelp_uri_get_path (uri), q_data->data);
+	}
 
 	g_free (command_line);
 
@@ -406,8 +303,29 @@ reader_man_info_start (ReaderThreadData *th_data)
 		g_signal_emit (reader, signals[ERROR], 0, error);
 		g_error_free (error);
 	} else {
-		g_async_queue_push (priv->thread_queue, q_data);
+		if (yelp_uri_get_type (uri) == YELP_URI_TYPE_DOCBOOK_XML ||
+		    yelp_uri_get_type (uri) == YELP_URI_TYPE_DOCBOOK_SGML) {
+			gchar *chunk;
+			
+			if (yelp_uri_get_section (uri) &&
+			    strcmp (yelp_uri_get_section (uri), "")) {
+				chunk = reader_get_chunk (q_data->data,
+							  yelp_uri_get_section (uri));
+			} else {
+				chunk = reader_get_chunk (q_data->data,
+							  "toc");
+			}
+
+			g_free (q_data->data);
+			q_data->data = chunk;
+
+			if (!q_data->data) {
+				q_data->data = g_strdup ("<html><body></body></html>");
+			}
+		}
 		
+		g_async_queue_push (priv->thread_queue, q_data);
+			
 		q_data = reader_q_data_new (reader, priv->stamp, 
 					    READER_QUEUE_TYPE_FINISHED);
 
@@ -428,8 +346,7 @@ reader_file_start (ReaderThreadData *th_data)
 	gchar             buffer[BUFFER_SIZE];
 	GnomeVFSFileSize  n;
 	
-	g_return_if_fail (YELP_IS_READER (reader));
-	g_return_if_fail (uri != NULL);
+	g_return_if_fail (th_data != NULL);
 
 	reader = th_data->reader;
 	priv   = reader->priv;
@@ -485,7 +402,7 @@ reader_file_start (ReaderThreadData *th_data)
 	}
 	
 	q_data = reader_q_data_new (reader, stamp, READER_QUEUE_TYPE_FINISHED);
-	
+
 	g_async_queue_push (priv->thread_queue, q_data);
 }
 
@@ -523,13 +440,17 @@ reader_start (ReaderThreadData *th_data)
 	
 	d(g_print ("reader_start\n"));
 
+	if (!yelp_uri_exists (uri)) {
+		return NULL;
+	}
+
 	STAMP_MUTEX_LOCK;
 	
 	if (reader_check_cancelled (reader, th_data->stamp)) {
 		STAMP_MUTEX_UNLOCK;
 
 		/* FIXME: refs??? */
-/* 		reader_th_data_free (th_data); */
+/* 		reader_th_data_free(th_data); */
 
 		return NULL;
 	}
@@ -541,13 +462,13 @@ reader_start (ReaderThreadData *th_data)
 	STAMP_MUTEX_UNLOCK;
 	
 	switch (yelp_uri_get_type (uri)) {
+/* 		reader_db_start (th_data); */
+/* 		break; */
 	case YELP_URI_TYPE_DOCBOOK_XML:
 	case YELP_URI_TYPE_DOCBOOK_SGML:
-		reader_db_start (th_data);
-		break;
 	case YELP_URI_TYPE_MAN:
 	case YELP_URI_TYPE_INFO:
-		reader_man_info_start (th_data);
+		reader_convert_start (th_data);
 		break;
 	case YELP_URI_TYPE_HTML:
 		reader_file_start (th_data);
@@ -555,9 +476,6 @@ reader_start (ReaderThreadData *th_data)
 	case YELP_URI_TYPE_TOC:
 		/* Should this be handled here?? */
 		break;
-	case YELP_URI_TYPE_NON_EXISTENT:
-		/* Set GError and signal error */
-		break;
 	case YELP_URI_TYPE_INDEX:
 	case YELP_URI_TYPE_PATH:
 		/* Hmm .. is this wrong or what?? */
@@ -664,7 +582,8 @@ reader_idle_check_queue (ReaderThreadData *th_data)
 				priv->stamp = 1;
 			}
 
-			g_signal_emit (reader, signals[FINISHED], 0);
+			g_signal_emit (reader, signals[FINISHED], 0,
+				       th_data->uri);
 
 			ret_val = FALSE;
 			break;
@@ -703,6 +622,7 @@ reader_q_data_free (ReaderQueueData *q_data)
 	g_free (q_data->data);
 	g_free (q_data);
 }
+
 #if 0
 static void
 reader_th_data_free (ReaderThreadData *th_data)
@@ -715,8 +635,78 @@ reader_th_data_free (ReaderThreadData *th_data)
 	g_free (th_data);
 }
 #endif
+
+static gchar *
+reader_get_chunk (const gchar *document, const gchar *section)
+{
+	gchar       *header;
+	gchar       *chunk;
+	const gchar *footer;
+	gchar       *ret_val;
+	const gchar *start;
+	const gchar *end;
+	gchar       *tag;
+	GTimer      *timer;
+	
+/* 	g_print ("%s\n", document); */
+
+	timer = g_timer_new ();
+
+	end = strstr (document, "<!-- End of header -->");
+	
+	if (!end) {
+/* 		g_warning ("Wrong type of document\n"); */
+		return g_strdup (document);
+	}
+	
+	header = g_strndup (document, end - document);
+	
+	tag = g_strdup_printf ("<!-- Start of chunk: [%s] -->", section);
+	start = strstr (document, tag);
+	g_free (tag);
+	
+	if (!start) {
+/* 		g_warning ("Document doesn't include section: '%s'", section); */
+		g_free (header);
+
+		return g_strdup (document);
+	}
+
+	end = strstr (start, "<!-- End of chunk -->");
+
+	if (!end) {
+/* 		g_warning ("Document is doesn't contain end tag for section: %s", */
+/* 			   section); */
+		g_free (header);
+
+		return g_strdup (document);
+	}
+	
+	chunk = g_strndup (start, end - start);
+	
+	footer = strstr (document, "<!-- Start of footer -->");
+	
+	if (!footer) {
+/* 		g_warning ("Couldn't find footer in document"); */
+		g_free (header);
+		g_free (chunk);
+
+		return g_strdup (document);
+	}
+	 
+	ret_val = g_strconcat (header, chunk, footer, NULL);
+	
+	g_free (header);
+	g_free (chunk);
+
+/* 	g_print ("Finding chunk took: %f seconds\n",  */
+/* 		 g_timer_elapsed (timer, 0)); */
+
+	return ret_val;
+}
+
 YelpReader *
-yelp_reader_new (gboolean async)
+yelp_reader_new ()
 {
         YelpReader     *reader;
         YelpReaderPriv *priv;
@@ -725,20 +715,18 @@ yelp_reader_new (gboolean async)
         
 	priv = reader->priv;
 	
-	priv->async = async;
-
         return reader;
 }
 
-void
+gboolean
 yelp_reader_start (YelpReader *reader, YelpURI *uri)
 {
         YelpReaderPriv   *priv;
 	ReaderThreadData *th_data;
 	gint              stamp;
 	
-	g_return_if_fail (YELP_IS_READER (reader));
-	g_return_if_fail (uri != NULL);
+	g_return_val_if_fail (YELP_IS_READER (reader), FALSE);
+	g_return_val_if_fail (uri != NULL, FALSE);
 
 	priv = reader->priv;
 
@@ -752,22 +740,62 @@ yelp_reader_start (YelpReader *reader, YelpURI *uri)
 	
 	stamp = priv->stamp;
 
+	STAMP_MUTEX_UNLOCK;
+
 	th_data = g_new0 (ReaderThreadData, 1);
 	th_data->reader = g_object_ref (reader);
 	th_data->uri    = yelp_uri_ref (uri);
 	th_data->stamp  = stamp;
 
-	STAMP_MUTEX_UNLOCK;
+	if (yelp_uri_get_type (uri) == YELP_URI_TYPE_DOCBOOK_XML || 
+	    yelp_uri_get_type (uri) == YELP_URI_TYPE_DOCBOOK_SGML) {
+		const gchar *document;
+		gchar *chunk;
 
-	if (priv->async) {
-		g_idle_add ((GSourceFunc) reader_idle_check_queue, th_data);
+		document = yelp_cache_lookup (yelp_uri_get_path (uri));
+		
+		if (document) {
+			if (yelp_uri_get_section (uri) &&
+			    strcmp (yelp_uri_get_section (uri), "")) {
+				chunk = reader_get_chunk (document, 
+							  yelp_uri_get_section (uri));
+			} else {
+				chunk = reader_get_chunk (document, "toc");
+			}
+		
+			if (chunk) {
+				ReaderQueueData *q_data;
+				
+				q_data = reader_q_data_new (reader, stamp, 
+							    READER_QUEUE_TYPE_START);
+				g_async_queue_push (priv->thread_queue, 
+						    q_data);
 
-		g_thread_create ((GThreadFunc) reader_start, th_data,
-				 TRUE, 
-				 NULL /* FIXME: check for errors */);
-	} else {
-		reader_start (th_data);
+				q_data = reader_q_data_new (reader, stamp, 
+							    READER_QUEUE_TYPE_DATA);
+
+				q_data->data = chunk;
+				g_async_queue_push (priv->thread_queue, 
+						    q_data);
+				
+				q_data = reader_q_data_new (reader, stamp, 
+							    READER_QUEUE_TYPE_FINISHED);
+
+				g_async_queue_push (priv->thread_queue, 
+						    q_data);
+
+				g_idle_add ((GSourceFunc) reader_idle_check_queue, th_data);
+				return TRUE;
+			}
+		}
 	}
+	
+	g_timeout_add (100, (GSourceFunc) reader_idle_check_queue, th_data);
+
+	g_thread_create ((GThreadFunc) reader_start, th_data,
+			 TRUE, 
+			 NULL /* FIXME: check for errors */);
+	return FALSE;
 }
 
 void
